/** =======================================================================
*** Net Image Filter -- a netpbm based image filtering service
*** 
*** by Marco Scheurer, Sen:te & Bill Bumgarner, Netsurfer, Inc.
*** 
*** Version 1.1
*** 
*** Copyright 1995, Sente Sarl. All rights reserved. Copyright 1995,
*** Netsurfer, Inc.. All Rights Reserved.
*** 
*** This notice may not be removed from this source code.
*** 
*** This object is included in the MiscKit by permission from the author(s)
*** and its use is governed by the MiscKit license, found in the file
*** "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
*** for a list of all applicable permissions and restrictions.
*** =======================================================================
**/

/* Version 1.1 by b.bum <bbum@friday.com>
   
   v1.1 is pretty much a complete re-write using v1.0 as a shell project.
 */

#import "NetImageConverter.h"
#import <appkit/appkit.h>

#import <mach/mach.h>

#import <signal.h>

#import <SurfImage/surfimage.h>

#define VERSION     "v1.1"
#define DEFAULTS_DB	"NetImageFilter"
#define DEBUG_DFLT  "debug"
#define STDERR_DFLT  "stderr"
#define ESCAPED_CHARS "'\"#{}[]() \t\n"

static NXDefaultsVector FilterDefaults = {
	{"jpeg", "/usr/local/bin"},
	{"netpbm", "/usr/local/imagetools"},
	{DEBUG_DFLT, "NO"},
	{STDERR_DFLT, "NO"},
	{NULL}
};

static const char *NoFilterFound = "No Filter Available";

/* Random places to look for the netpbm/jpeg binaries.  Feel free
   to add new paths -- invalid paths are automagically eliminated.
 */
static const char *BinPaths[] = {
	"~/Unix/bin",
	"~/Unix/netpbm",
	"~/Unix/netpbm/bin",
	"~/Unix/imagetools",
	"~/Unix/imagetools/bin",
	"/LocalDeveloper/bin",
	"/LocalDeveloper/netpbm",
	"/LocalDeveloper/netpbm/bin",
	"/LocalDeveloper/imagetools",
	"/LocalDeveloper/imagetools/bin",
	"/LocalApps",
	"/LocalApps/netpbm",
	"/LocalApps/netpbm/bin",
	"/LocalApps/imagetools",
	"/LocalApps/imagetools/bin",
	"/usr/local/bin",
	"/usr/local/netpbm",
	"/usr/local/netpbm/bin",
	"/usr/local/imagetools",
	"/usr/local/imagetools/bin",
	NULL	
};

@implementation NetImageConverter
/*"
 * Implements a generic image filtering service withn internal support for
 * JPEG and GIF (including transparency) via the SurfImage library.  The
 * SurfImage library is the same decoder used in the rather awesome
 * Netsurfer Internet Resource Navigator (from Netsurfer, Inc;
 * http://www.netsurfer.com).
 * 
 * For images not directly converted by SurfImage, NetImageFilter searches
 * high and low for the appropriate netpbm tool used to convert the
 * particular image format to a ppm image.  From the ppm, #pnmtotiff is
 * used to convert the image to a tiff.
 * 
 * Sending the NetImageFilter process a SIGUSR2 will cause the internal
 * filter path cache to be emptied.  SIGUSR1 will print all of the entries
 * currently in the cache.
"*/

+ sharedInstance
/*"
 * Returns a single shared instance of #NetImageConverter.
"*/
{
	return [self sharedInstance: NO];
}

+ sharedInstance:(BOOL) debugFlag
/*"
 * Returns a single shared instance of #NetImageConverter.  Upon the first
 * invocation (the one that allocates the shared instance), if debugFlag is
 * YES, then the instance will be allocated with debugging enabled.
"*/
{
	static id sharedInstance = nil;

	if (!sharedInstance) {
		sharedInstance = [self alloc];
		[sharedInstance setDebug:debugFlag];
		[sharedInstance init];
	}

	return sharedInstance;
}

void handle_usr1 () {
	id nic = [NetImageConverter sharedInstance];
	[nic printCache];
}

void handle_usr2 () {
	id nic = [NetImageConverter sharedInstance];
	[nic resetCache];
}

- init
/*"
 * Designated initializer.  Reads defaults from the defaults database and
 * configures the filter appropriately.
"*/
{
	const char **paths;
	const char *dValue;
	
    /* register defaults vector
     */
    NXRegisterDefaults (DEFAULTS_DB, FilterDefaults);

	signal (SIGUSR1, handle_usr1);
	signal (SIGUSR2, handle_usr2);

	NXLogError("(NetImageFilter %s) [Initialization]\n", VERSION);

	if((![self isDebug]) &&
	   (dValue = NXGetDefaultValue(DEFAULTS_DB, DEBUG_DFLT)) &&
	   ( (dValue[0] == 'y') || (dValue[0] == 'Y') || (dValue[0] == '1') ) )
		[self setDebug: YES];

	if((dValue = NXGetDefaultValue(DEFAULTS_DB, STDERR_DFLT)) &&
	   ( (dValue[0] == 'y') || (dValue[0] == 'Y') || (dValue[0] == '1') ) )
		[self setSpewSTDERR: YES];

    /* Create binaryHash -- keys and values are strings [that is,
       we manage all the memory for the keys/values stored in the
       table].
     */
	binaryHash = [[HashTable allocFromZone:[self zone]]
					  initKeyDesc: "*" valueDesc: "*"];

	binPaths = [[List allocFromZone:[self zone]] init];
	paths = BinPaths;

	while (*paths) {
		id bundle;
		id vBundle;

        // -initForDirectory: may not return the id of the bundle that was
        // originally called.  If some other bundle already exists for the
        // path, -initForDirectory: returns the other bundle.  To avoid
        // duplicates, we detect this and throw out any duplicates.

		bundle = [NXBundle allocFromZone:[self zone]];
		vBundle = [bundle initForDirectory: *paths];

		if(vBundle == bundle)
			[binPaths addObject:bundle];

		paths++;
	}

#ifdef SNOWBALL_IN_HELL
    /* This is only necessary when using popen()
     */
	dataBuf = (char *) malloc(vm_page_size);
#endif

	decoder = [[SurfImageDecoder allocFromZone:[self zone]] init];
	if(isDebug)
		[decoder setVerboseMode:YES];

	return self;
}

- setDebug:(BOOL) aFlag
/*"
 * Enables and disables debugging messages.
"*/
{
	isDebug = aFlag;
	return self;
}

- (BOOL) isDebug
/*"
 *  Returns YES if debugging messages are enabled.
"*/
{ return isDebug; }

- setSpewSTDERR:(BOOL) aFlag
/*"
 * Enables and disables spewage of subprocess's stderr to stderr.
"*/
{ spewStderr = aFlag; return self; }

- (BOOL) spewSTDERR
/*"
 * Returns YES if subprocesses stderr will be spewed to stderr.
"*/
{ return spewStderr; }

void free_key(void * key)
{
	free(key);
}

void free_value(void * value)
{
	if(value && (value != NoFilterFound))
		free(value);
}

- resetCache
/*"
 * Resets the cached hashtable of decoder paths.  Useful if one has moved
 * around various image decoding binaries and the image filter needs to
 * look for the binaries again.
 * 
 * A cache reset can be triggered by sending the NetImageFilter process a
 * SIGUSR2.
"*/
{
	const char *dValue;
	
	if((dValue = NXGetDefaultValue(DEFAULTS_DB, STDERR_DFLT)) &&
	   ( (dValue[0] == 'y') || (dValue[0] == 'Y') || (dValue[0] == '1') ) )
		[self setSpewSTDERR: YES];

//	fprintf(stderr, "NetImageFilter (%s) [Reset Cache]\n", VERSION);
	NXLogError("(NetImageFilter %s) [Reset Cache]\n", VERSION);

	[binaryHash freeKeys:free_key values:free_value];

	return self;
}

- printCache
/*"
 * Prints the content of the binary path cache.  A cache print can also be
 * triggered by sending the NetImageFilter process a SIGUSR1.
"*/
{
	NXHashState aState;
	const char *key;
	const char *value;
	
	NXLogError("(NetImageFilter %s) [Print Cache]\n", VERSION);

	NXLogError("(NetImageFilter %s) [%d entries]\n",
			VERSION,
			[binaryHash count]);	

	aState = [binaryHash initState];
	while([binaryHash nextState:&aState
					  key:(void **) &key
					  value:(void **) &value])
		NXLogError("(NetImageFilter %s) [Cache entry]: {%s} -> {%s}\n",
				   VERSION,
				   key,
				   value);

	return self;
}

- (const char *) filterPathFromData: (const char *) userData
/*"
 * Converts the name and family of a converter (see the serrvices file)
 * into a path to an executable used to convert that particular type of
 * image.  Returns NULL if no filter found.
"*/
{
	const char *filterPath;
	char *separator;
	char *userDataMunge;
	char *converter;
	char *family;
	
/* BTW:  MAXPATHLEN is safe to use without the traditional "+1" --
   a friend tested out the library routines and found that they all
   assume that MAXPATHLEN chars INCLUDES the terminating NULL!

   b.bum
 */

	if (!userData) return NULL; // validate argument.

    /* Check to see if the path to the filter program described by
       userData is already hashed in binaryHash.  If it is, just
       return that path.  Note if the empty string (a string of
       length 0) is returned by valueForKey, it indicates that a
       search has already been performed, but the binary in question
       was not found.
     */
	filterPath = (const char *) [binaryHash valueForKey: userData];

    /* If a filterPath is returned, return it unless the NoFilterFound
       string is returned.  If NoFilterFound is returned, then return
       NULL.
     */
	if (filterPath)
		return (filterPath != NoFilterFound) ? filterPath : NULL;

    /* Make a mungeable copy of userData
     */
	userDataMunge = NXCopyStringBufferFromZone(userData, [self zone]);

    /* Search for the binary.
     */

    /* Determine whether or not a family name was specified.  If it
       was, then split the family name from the converter.
     */
	separator = index (userDataMunge, '/');
	if (separator && *separator) {
        /* kill separator
         */
		separator[0] = '\0';
		converter	 = separator + 1;
		family		 = userDataMunge;
	} else {
		converter	 = userDataMunge;
		family		 = NULL;
	}

    /* 1. Look for path in defaults database using "converter" as the
       key
     */
	filterPath = [self validateConverterPath:
					   NXGetDefaultValue(DEFAULTS_DB, converter)];

    /* 2. [if not already found] If the family has been specified,
       look for the binary rooted at the value of the default value
       associated with the family name.
     */
	if (!filterPath && family)
		filterPath = [self validateConverter:converter
						   inDirectory:NXGetDefaultValue(DEFAULTS_DB, family)];

   /*
    * 3. special case 'cat' such that there is no need to do popen()
    */
	if (!strcmp(userData, "cat"))
		filterPath = [self validateConverter:"cat"
						   inDirectory:"/bin"];

    /* 3. [if not already found] run popen() and use return value
     */
	if (!filterPath) {
		char cmdBuf[1024];
		FILE *process;

		sprintf(cmdBuf, "which %s\n", converter);

		process = popen(cmdBuf, "r");

		if (process) {
			int fd = fileno(process);
			int count;
			char buf[MAXPATHLEN];

			count = read(fd, buf, MAXPATHLEN);

			if (count && (buf[0] == '/')) {
				char *index = index(buf, '\n');
				
				if(index)
					*index = '\0';

				filterPath = [self validateConverterPath: buf];
			}
			
			pclose(process);
		}
	}

    /* 4. check list of standard paths.
     */
	if (!filterPath) {
		id *ida, *max;

		ida = NX_ADDRESS(binPaths);
		max = ida + [binPaths count];

		for(;ida<max && !filterPath;ida++)
			filterPath = [self validateConverter:converter inBundle:*ida];
	}

    /* 5. Check bundleForClass:[self class] and "netpbm", "netpbm/bin"
       directories
     */
	if (!filterPath) {
		id bundle = [NXBundle bundleForClass:[self class]];
		
		filterPath = [self validateConverter:converter inBundle:bundle];

		if (!filterPath) {
			char pathBuf[MAXPATHLEN];

			sprintf(pathBuf, "%s/netpbm", [bundle directory]);
			filterPath = [self validateConverter:converter
							   inDirectory:pathBuf];

			if(!filterPath) {
				strcat(pathBuf, "bin");
				filterPath = [self validateConverter:converter
								   inDirectory:pathBuf];
			}
		}
	}

    /* Cache found path (or cache NoFilterFound string).
     */
	[binaryHash insertKey:NXCopyStringBufferFromZone(userData, [self zone])
				value:(void *) (filterPath ? filterPath : NoFilterFound)];

    /* clean up
     */
	free(userDataMunge);

	return filterPath;
}

- (const char *) validateConverter:(const char *) aConverter
                          inBundle:aBundle
/*"
 * Validates the converter aConverter in bundle aBundle.  Returns the path
 * to the converter if the specified converter exists and is executable.
"*/
{
	char pathBuf[MAXPATHLEN];

	if (!aConverter || !aBundle)
		return NULL;

	if([aBundle getPath:pathBuf forResource:aConverter ofType:NULL])
		return [self validateConverterPath:pathBuf];

	return NULL;
}

- (const char *) validateConverter:(const char *) aConverter
                       inDirectory:(const char *) aDirPath
/*"
 * Returns the path to aConverter contained in aDirPath if aConverter
 * exists and is executable.
"*/
{
	char pathBuf[MAXPATHLEN];

	if(!aConverter || !aDirPath)
		return NULL;

    /* Technically, this could fail if aConverter+aDirPath is longer
       than MAXPATHLEN.  That isn't very likely.  I'm choosing to
       ignore that possibility in the hopes that many other things
       will break first.
     */
	sprintf(pathBuf, "%s/%s", aDirPath, aConverter);

	return [self validateConverterPath: pathBuf];
}

- (const char *) validateConverterPath:(const char *) aPotentialPath
/*"
 * Returns a copy of aPotentialPath if aPotentialPath is a path to an
 * exectuable file.
"*/
{
	struct stat statBuf;
	
	if (!aPotentialPath)
		return NULL; // no path specified

	if(stat(aPotentialPath, &statBuf) == -1)
		return NULL; // could not stat

	if(statBuf.st_mode & S_IFDIR)
		return NULL; // is a directory

    /* check file read and exec attributes
     */
	if( (statBuf.st_mode & S_IREAD) && (statBuf.st_mode & S_IEXEC) )
        /* If the file is both readable and executable, then make
           a copy of the path and return the newly created string.
         */
		return NXCopyStringBufferFromZone(aPotentialPath,
										  [self zone]);

	return NULL; // by default, fail.
}

/* This is a hacque -- i have found problems w/the various NeXT-included
   temporaryy file name production functions.  since generating a
   unique filename is such a trivial task [within the scope of NIF's
   problem space], i decided to simply provide the following hacque.
 */
- (const char *) _tempPath
/*"
 * Generates a temporary path with a ".tiff" extension in /tmp/.  Uses the
 * process ID of the filter and a generic counter to generate the unique
 * path.
"*/
{
	static char tempPath[MAXPATHLEN];
	static unsigned int number = 0;
	static int  pid	   = 0;

	if(!pid)
		pid = getpid();

	sprintf(tempPath, "/tmp/NiFtMp_%d_%d.tiff", pid, number);

	number++;

	return tempPath;
}

#define MATCHBUFSIZE 256
const unsigned char *Expand_Match_Description(const unsigned char *desc)
{
	static unsigned char matchBuffer[MATCHBUFSIZE];
	register unsigned char *s, *sEnd, cur_char;
	register short i;
	BOOL lNot=NO;

	// set up the beginning and end pointer
	s   = (unsigned char *) desc;
	sEnd = s + strlen((char *) desc);

	// if the first character is '^', then set the inversion flag and advance s
	if( (s != sEnd) && (*s == '^')) {
		lNot=YES;
		s++;
	}
  
	// clear last result
	bzero(matchBuffer, sizeof(matchBuffer));

	// traverse expression filling table w/marks as we go.
	while(s!=sEnd) {
		cur_char = *s++;
		
		// check for an '/' escape
		if(cur_char == '/'){
			// if escape, but at end, then break
			if(s == sEnd) break;
			// else, skip escape char and move to next char
			cur_char = *s++;
		
			switch(cur_char){
// <<CHANGE>> need to add other equivalents here
				case 'n' :
					cur_char='\n';
					break;
				case 't' :
					cur_char='\t';
					break;
			}
		}
		
		// check for '-' sequence
		if( (s != sEnd) && (*s == '-')){
			// if end follows the '-', then break
			if( (++s) == sEnd)
				break;
			// have final char, loop incrementing cur_char until > *s
			while(cur_char <= *s)
				matchBuffer[cur_char++]=1;
			// increment past final char
			s++;
		} else
			// single char, nothing special, mark and continue
			matchBuffer[cur_char]=1;
	}

	// use exclusive-or to negate the matchBuffer, if the first
	// character was '^'
	if(lNot)
		for(i=0;i<sizeof(matchBuffer);i++)
			matchBuffer[i] ^= 1;

	return matchBuffer;
}

- (const char *) _sourcePathFrom:(const char *) path length:(int) len
{
	static char returnPath[MAXPATHLEN];
	int rpIndex, sIndex;
	static const char *escapeBuf;

	if(!escapeBuf) {
		const char *matchBuf = Expand_Match_Description(ESCAPED_CHARS);
		escapeBuf = (const char *) malloc(MATCHBUFSIZE);
		if(!matchBuf) {
			fprintf(stderr, "COULD NOT EXPAND MATCH BUFFER "
					"-- DEATH ENSUES!\n");
			exit(1);
		}
		bcopy(matchBuf, (char *)escapeBuf, MATCHBUFSIZE);
	}

	for(rpIndex=0, sIndex = 0;
		(rpIndex < MAXPATHLEN) && (sIndex < len);
		rpIndex++, sIndex++) {

		if(escapeBuf[path[sIndex]]) {
			returnPath[rpIndex++] = '\\';
			if(rpIndex == MAXPATHLEN)
				return NULL;
		}

		returnPath[rpIndex] = path[sIndex];
				
	}

	if(rpIndex == MAXPATHLEN)
		return NULL;

	returnPath[rpIndex] = '\0';

	return returnPath;	
}

#ifndef SNOWBALL_IN_HELL
- anyToTIFF:pasteboard
   userData:(const char *)userData
	  error:(char **)errorMessage
/*"
 * Converts the image in the file contained on pasteboard to a tiff and
 * writes the resulting tiff back to the pasteboard.  userData contains a
 * description of the image filter to use in the form of "family/filter".
 * An extensive search is made for each image filter.  Once an image filter
 * is identified, the path to the filter is cached.
"*/
{
    NXAtom  foundType = [pasteboard types][0];

	if(isDebug)
		NXLogError("(NetImageFilter %s): Received Request {%s}\n",
				   VERSION,
				   userData);	

	if (foundType) {
		char    *data  = NULL;
		int     length = 0;
		id      image;
        
		[pasteboard readType:foundType data:&data length:&length];

		if(isDebug)
			NXLogError("(NetImageFilter %s): Read Pasteboard {%d bytes}\n",
					VERSION,
					length);

		if(decoder) 
			image = [decoder decodeFromFile:data];
		else
			image = nil;

		if(image) {
			NXStream *tiffStream = NXOpenMemory(NULL, 0, NX_READWRITE);

			if(isDebug)
				NXLogError("(NetImageFilter %s): Used SurfDeecoder\n",
					VERSION);

			[image writeTIFF: tiffStream allRepresentations: NO];
			
			[pasteboard declareTypes:&NXTIFFPboardType num:1 owner:self];
			[pasteboard writeType:NXTIFFPboardType fromStream:tiffStream];

			[image free];
			NXCloseMemory (tiffStream, NX_FREEBUFFER);
		} else if (data && length && userData && *userData) {
			char commandFormat[] = "%s %s | %s -none > %s\n";
			char command [(4 * MAXPATHLEN) + strlen (commandFormat)];

			const char *sourcePath;
			const char *tiffPath;
			const char *toPNMConverter;
			const char *toTIFFConverter;

			int status;
			NXStream *tiffStream;

			sourcePath		= [self _sourcePathFrom:data length:length];
			toPNMConverter	= [self filterPathFromData:userData];
			toTIFFConverter	= [self filterPathFromData:"netpbm/pnmtotiff"];

			if (!toPNMConverter || !toTIFFConverter) {
				NXLogError("(NetImageFilter %s): Could not find converter "
						   "for request {%s}; {%s} -> {%s}\n",
						   VERSION,
						   userData,
						   toPNMConverter,
						   toTIFFConverter);
				return nil;
			}

			tiffPath = [self _tempPath];

			sprintf (command, commandFormat,
					 toPNMConverter,
					 sourcePath,
					 toTIFFConverter,
					 tiffPath);

			if	(isDebug)
				NXLogError("(NetImageFilter %s): Conversion command {%s}\n",
						   VERSION,
						   command);

			if(!spewStderr)
				freopen ("/dev/null", "w", stderr);

			status = system(command);

			if(!spewStderr)
				freopen ("/tmp/console.log", "a", stderr);
			
			if (status) {
				NXLogError("(NetImageFilter %s): Conversion failed; "
						   "exit code (%d).  Command was: {%s}\n",
						   VERSION	,
						   status,
						   command);
				/* unlink temporary file
				 */
				unlink(tiffPath);
				return nil;
			}

            /* write TIFF to PasteBoard
             */
			tiffStream = NXMapFile(tiffPath, NX_READONLY);

			if	(isDebug) {
				char *buf;
				int  length, max;

				NXGetMemoryBuffer(tiffStream, &buf, &length, &max);
				NXLogError("(NetImageFilter %s): Writing to "
						   "Pasteboard {%d bytes}\n",
						   VERSION,
						   length);
			}

			[pasteboard declareTypes:&NXTIFFPboardType num:1 owner:self];
			[pasteboard writeType:NXTIFFPboardType fromStream:tiffStream];

			NXCloseMemory (tiffStream, NX_FREEBUFFER);

            /* unlink temporary file
             */
			unlink(tiffPath);
		}
	}

	return self;
}
#endif

#ifdef SNOWBALL_IN_HELL
/*
 * The following code implements the conversion using popen().  Itworks
 * perfectly except for one minor snafu-- pnmtotiff HAS to be able to
 * perform seek() operations on the output FD;  in other words, it can spew
 * at a file, but it can't spew into a pipe...
 * 
 * ppm2tiff (libtiff's rendition of pnmtotiff) requires that one specify a
 * file name..
 * 
 * If anyone cares to fix these problems, it would make the converter run a
 * bunch faster!
 * 
 * If you decide to fix this method such that it works-- it needs to be
 * renamed to -anyToTIFF:userData:error: (or change the services file) and
 * you need to add the support for using SurfImageDecoder.
 * 
 * b.bum <bbum@friday.com>
 */
- _anyToTIFF:pasteboard
   userData:(const char *)userData
	  error:(char **)errorMessage
{
    NXAtom  foundType = [pasteboard types][0];

	fprintf(stderr,
			"NetImageFilter v1.1: Received Request {%s}\n",
			userData);	

	if (foundType) {
		char    *data = NULL;
		int     length = 0;
        
		[pasteboard readType:foundType data:&data length:&length];

		fprintf(stderr,
				"NetImageFilter v1.1: Read Pasteboard {%d bytes}\n",
				length);

		if (data && length && userData && *userData) {
			char commandFormat[] = "%s %s | %s -none\n";
			char command [(3 * MAXPATHLEN) + strlen (commandFormat)];
			char sourcePath [MAXPATHLEN];

			const char *toPNMConverter;
			const char *toTIFFConverter;

			FILE *process;

			strncpy (sourcePath, data, length);
			sourcePath[length] = 0;

			toPNMConverter	= [self filterPathFromData:userData];
			toTIFFConverter	= [self filterPathFromData:"netpbm/pnmtotiff"];

			fprintf(stderr, "%s -> %s\n", toPNMConverter, toTIFFConverter);

			if(!toPNMConverter || !toTIFFConverter)
				return self;

			sprintf (command, commandFormat,
					 toPNMConverter, sourcePath, toTIFFConverter);

			fprintf(stderr, "Conversion Command: %s", command);

			process = popen(command, "r");

			if (process) {
				NXStream *tiffStream = NXOpenMemory(NULL, 0, NX_READWRITE);
				size_t bytes_read;

				while (bytes_read = fread(dataBuf, 1, vm_page_size, process)) {
					fprintf(stderr, "Writing {%d} bytes...\n", bytes_read);
					NXWrite(tiffStream, dataBuf, bytes_read);
				}

				fprintf(stderr, "Writing Tiff\n");

				[pasteboard declareTypes:&NXTIFFPboardType num:1 owner:self];
				NXSeek(tiffStream, 0, NX_FROMSTART);
				[pasteboard writeType:NXTIFFPboardType fromStream:tiffStream];
				NXCloseMemory (tiffStream, NX_FREEBUFFER);
				
				pclose(process);
			}
		}
	}

	return self;
}
#endif
@end
